% The FourierSlice4D3 function was written by Zhan Yu 
%(https://www.eecis.udel.edu/~zyu/) for the purposes of implementation 
% of light field refocusing via Fourier slicing. Some modifications have 
% been made by me.
%
% The FDWF script was written by me based on a paper by Sari et. al.,
% cited in the final report

clear, clc, close all

% Gaussian interpolation radius
radius = 2;

% Load images
 I1 = imread('2.jpg');
 I2 = imread('1.jpg');
%I1 = imread('3new.jpg');
%I2 = imread('2new.jpg');
I1 = im2double(I1);
I2 = im2double(I2);

[height width layer] = size(I1);

% Take image difference and delta-z
Idiff = I1 - I2;
Idiff = Idiff - min(Idiff(:)); % normalize (test)
Idiff = Idiff./max(Idiff(:));
dz = 1.5e-3; % mm

% Calculate spatial frequencies
Fs_x = width/7.18; % pixels/mm
Fs_y = height/5.32;
dx = 1./Fs_x; % mm/pixel
dy = 1./Fs_y; 
[P,Q,~] = size(I1); % pixels
x = dx*(0:Q-1);    % mm
y = dy*(0:P-1); 
dFx = Fs_x/Q;    % cycles/px
dFy = Fs_y/P;
Fx = (-Fs_x/2:dFx:Fs_x/2-dFx);     % cycles/mm
Fy = (-Fs_y/2:dFy:Fs_y/2-dFy);

% Create filter matrix
H = zeros(P,Q);
for i = 1:P
    for j = 1:Q
        if (Fx(1,j)^2 + Fy(1,i)^2 == 0)
            H(i,j) = 1;
        else
            H(i,j) = 1./(-4*pi^2.*(Fx(1,j)^2 + Fy(1,i)^2));
        end
    end
end

% Noise correction, uncomment as needed
for i = 1:3
    I(:,:,i) = fftshift(fft2(ifftshift(Idiff(:,:,i))));
    %J(:,:,i) = I(:,:,i);
    
%     tic
%     % Ideal
%     if i == 3   % account for blue noise
%         co = 8.5;
%     else
%         co = 1.6;
%     end
%     C = ideal(I(:,:,i),co,395);
%     J(:,:,i) = C.*I(:,:,i);
%     toc
    
%     tic
%     % Gaussian
%     if i == 3   % account for blue noise
%         co = 8;
%     else
%         co = 1.1;
%     end
%     G = (1-gauss(I(:,:,i),co) + gauss(I(:,:,i),200))./2;
%     G = G - min(G(:)); G = G./max(G(:));
%     J(:,:,i) = G.*I(:,:,i);
%     toc
       
%     tic
%     % Butterworth
%     if i == 3   % account for blue noise
%         co = 9;
%     else
%         co = 1.5;
%     end
%     B = (1-butterworth(I(:,:,i),6,co) + butterworth(I(:,:,i),6,300))./2;
%     B = B - min(B(:)); B = B./max(B(:));
%     J(:,:,i) = B.*I(:,:,i);
%     toc
    
%     tic
%     % FDWF
%     F1(:,:,i) = fftshift(fft2(ifftshift(Idiff(:,:,i))));
%     F2(:,:,i) = fft2(ifftshift(Idiff(:,:,i)));
%     F = (FDWF(F1(:,:,i),5,32) + FDWF(F2(:,:,i),2,32))./2;
%     F = F - min(F(:)); F = F./max(F(:));
%     J(:,:,i) = F.*I(:,:,i);
%     toc
    
    N(:,:,i) = fftshift(ifft2(ifftshift(J(:,:,i))));
end

% Calculate U
for i = 1:3
    X(:,:,i) = fftshift(fft2(ifftshift(N(:,:,i)/dz)));
    U(:,:,i) = fftshift(ifft2(ifftshift(H.*X(:,:,i))));
end

% imshow(mat2gray(log(abs(X)+1)))
% figure;
% imshow(mat2gray(log(abs(Y)+1)))
% figure;

% Calculate M
[MX,MY] = gradient(U);

for k = 1:3
    MX(:,:,k) = MX(:,:,k)./((I1(:,:,k)+I2(:,:,k))/2);
    MY(:,:,k) = MY(:,:,k)./((I1(:,:,k)+I2(:,:,k))/2);
end

angle1 = 0.183;

% Set angles
thetay = -tan(angle1):tan(angle1)/6:tan(angle1);
thetax = zeros(1,length(thetay));

divFact = 1;

% Numerical aperture of scope, normally 0.82 -- 0.12
sigma = 0.32;

[r,c,k] = size(MX);

L = zeros(r,c,length(thetax),length(thetay),3);

% Calculate L
for tx = 1:length(thetax)
    for ty = 1:length(thetay)
        for k = 1:3
                L(:,:,tx,ty,k) = I2(:,:,k).*exp(-(thetax(tx)-...
                    MX(:,:,k)).^2./sigma.^2 - ...
                    (thetay(ty)-MY(:,:,k)).^2./sigma.^2);
        end
    end
end

% L = 0.21*L(:,:,:,:,1)+0.72*L(:,:,:,:,2)+0.07*L(:,:,:,:,3);

% Create output photos. Figure 1 is the original image used for PSNR
% comparison, Figure 2 is the filtered image
L(isnan(L)) = 0;
I3 = abs(squeeze(L(:,:,5,5,:)));
imshow(I2)
print -depsc imNoNoise
figure;
imshow(I3)
print -depsc imNoise 
psnr(rgb2gray(I3),rgb2gray(I2))

% for k = 1:3
%     I3(:,:,k) = FourierSlice4D3(L(:,:,:,:,k),thetay(5),640,480,radius);
% end
% I3 = I3./max(I3(:));
% I3 = abs(I3);
% imshow(I2)
% print -depsc imNoNoise
% figure;
% imshow(I3)
% print -depsc imNoise 
% psnr(rgb2gray(I3),rgb2gray(I2))

% for k = 1:3
%     M1(:,:,k) = FourierSlice4D3(L(:,:,:,:,k),thetay(3)/divFact,640,480,radius);
%     M2(:,:,k) = FourierSlice4D3(L(:,:,:,:,k),thetay(4)/divFact,640,480,radius);
%     M3(:,:,k) = FourierSlice4D3(L(:,:,:,:,k),thetay(5)/divFact,640,480,radius);
%     M4(:,:,k) = FourierSlice4D3(L(:,:,:,:,k),thetay(6)/divFact,640,480,radius);
%     M5(:,:,k) = FourierSlice4D3(L(:,:,:,:,k),thetay(7)/divFact,640,480,radius);
%     M6(:,:,k) = FourierSlice4D3(L(:,:,:,:,k),thetay(8)/divFact,640,480,radius);
%     M7(:,:,k) = FourierSlice4D3(L(:,:,:,:,k),thetay(9)/divFact,640,480,radius);
%     M8(:,:,k) = FourierSlice4D3(L(:,:,:,:,k),thetay(10)/divFact,640,480,radius);
%     M9(:,:,k) = FourierSlice4D3(L(:,:,:,:,k),thetay(11)/divFact,640,480,radius);
% end
% 
% % M1 = FourierSlice4D3(L,-tan(0.183)/divFact,640,480,radius);
% % M2 = FourierSlice4D3(L,-tan(0.183)/2/divFact,640,480,radius);
% % M3 = FourierSlice4D3(L,tan(0),640,480,radius);
% % M4 = FourierSlice4D3(L,tan(0.183)/2/divFact,640,480,radius);
% % M5 = FourierSlice4D3(L,tan(0.183)/divFact,640,480,radius);
% 
% M1 = abs(M1);
% M1 = M1./max(M1(:));
% M2 = abs(M2);
% M2 = M2./max(M2(:));
% M3 = abs(M3);
% M3 = M3./max(M3(:));
% M4 = abs(M4);
% M4 = M4./max(M4(:));
% M5 = abs(M5);
% M5 = M5./max(M5(:));
% M6 = abs(M6);
% M6 = M6./max(M6(:));
% M7 = abs(M7);
% M7 = M7./max(M7(:));
% M8 = abs(M8);
% M8 = M8./max(M8(:));
% M9 = abs(M9);
% M9 = M9./max(M9(:));
% 
% pTime = 0.01;
% 
% for k = 1:100
%     imshow(M1,[])
%     pause(pTime);
%     imshow(M2,[])
%     pause(pTime);
%     imshow(M3,[])
%     pause(pTime);
%     imshow(M4,[])
%     pause(pTime);
%     imshow(M5,[])
%     pause(pTime);
%     imshow(M6,[])
%     pause(pTime);
%     imshow(M7,[])
%     pause(pTime);
%     imshow(M8,[])
%     pause(pTime);
%     imshow(M9,[])
%      pause(0.3);
%     
%     imshow(M8,[])
%     pause(pTime);
%     imshow(M7,[])
%     pause(pTime);
%     imshow(M6,[])
%     pause(pTime);
%     imshow(M5,[])
%     pause(pTime);
%     imshow(M4,[])
%     pause(pTime);
%     imshow(M3,[])
%     pause(pTime);
%     imshow(M2,[])
%     pause(pTime);
% end

% figure;
% for tx = 1:length(thetax)
%     for ty = 1:length(thetay)
%         for k = 1:3
%             M(:,:,k) = squeeze(L(:,:,tx,ty,k));
%             M(:,:,k) = abs(M(:,:,k));
%             M(:,:,k) = M(:,:,k)./max(M(:));
%         end
%         imshow(M)
%     end
% end

% I3 = abs(squeeze(L(:,:,1,2,:)));
% imshow(I2)
% print -depsc imNoNoise
% imshow(I3)
% print -depsc imNoise 
% psnr(rgb2gray(I3),rgb2gray(I2))

